//
// gjr(a, n) inverts the (symmetric) n by n square matrix
// pointed to by a.  It uses the Gauss-Jordan-Rutishauser
// algorithm with double pivoting.  Returns 1 if matrix is
// singular, 0 if not.  Essentially copied from nllsq.
// note: the matrix need not be symmetric for gjr to work.
//

#include "fit.h"

int gjr( dbl** a, int npar )
{
    array ar;
    dbl* c = ar.alloc_d(npar);
    dbl* d = ar.alloc_d(npar);
    dbl pivot,
        z,
        zeta = DBL_EPSILON;          // machine eps for sing test
    int* q = ar.alloc_i(npar);
    int* r = ar.alloc_i(npar);
    int i, j, k, m, p, sing = no;



    for ( k = 0 ; k < npar ; k++ ) {

// determination of pivot element
        pivot = 0.0;
        for (i = k ; i < npar ; i++)
            for (j = k ; j < npar ; j++) {
                if (fabs(a[i][j]) > fabs(pivot)) {
                    pivot = a[i][j];
                    r[k] = i;
                    q[k] = j;
                }
            }
        if (fabs(pivot) <= zeta) {
            sing = yes;
            return sing;
        }

// exchange of pivotal row with nth row

        if (r[k] != k)
            for (j = 0 ; j < npar ; j++) {
                p = r[k];
                z = a[p][j];
                a[p][j] = a[k][j];
                a[k][j] = z;
            }

// exchange of column

        if (q[k] != k)
            for (i = 0 ; i < npar ; i++) {
                p = q[k];
                z = a[i][p];
                a[i][p] = a[i][k];
                a[i][k] = z;
            }

// jordan step

        for (j = 0 ; j < npar ; j++) {
            if (j == k) {
                d[j] = 1.0/pivot;
                c[j] = 1.0;
            }
            else {
                d[j] = - a[k][j]/pivot;
                c[j] = a[j][k];
            }
            a[k][j] = 0.0;
            a[j][k] = 0.0;
        }
        for (i = 0 ; i < npar ; i++)
            for (j = 0 ; j < npar ; j++)
                a[i][j] += c[i]*d[j];
    }

// reorder the matrix

    for (m = 0 ; m < npar ; m++) {
        k = npar - 1 - m;
        if (r[k] != k)
            for (i = 0 ; i < npar ; i++) {
                p = r[k];
                z = a[i][p];
                a[i][p] = a[i][k];
                a[i][k] = z;
            }
        if (q[k] != k)
            for (j = 0 ; j < npar ; j++) {
                p = q[k];
                z = a[p][j];
                a[p][j] = a[k][j];
                a[k][j] = z;
            }
    }
    ar.dele(c);
    ar.dele(d);
    ar.dele(q);
    ar.dele(r);
    return sing ;
}

